/* Installation Info */
static const char qt_configure_prefix_path_str  [12+256] = "qt_prfxpath=/usr";
#ifdef QT_BUILD_QMAKE
static const char qt_configure_ext_prefix_path_str   [12+256] = "qt_epfxpath=/usr";
static const char qt_configure_host_prefix_path_str  [12+256] = "qt_hpfxpath=/usr";
#endif
static const short qt_configure_str_offsets[] = {
    0,
    14,
    26,
    30,
    46,
    58,
    74,
    90,
    102,
    110,
    120,
    143,
    162,
#ifdef QT_BUILD_QMAKE
    178,
    179,
    185,
    197,
    201,
    209,
    219,
#endif
};
static const char qt_configure_strs[] =
    "share/doc/qt5\0"
    "include/qt5\0"
    "lib\0"
    "lib/qt5/libexec\0"
    "lib/qt5/bin\0"
    "lib/qt5/plugins\0"
    "lib/qt5/imports\0"
    "lib/qt5/qml\0"
    "lib/qt5\0"
    "share/qt5\0"
    "share/qt5/translations\0"
    "share/qt5/examples\0"
    "share/qt5/tests\0"
#ifdef QT_BUILD_QMAKE
    "\0"
    "false\0"
    "lib/qt5/bin\0"
    "lib\0"
    "lib/qt5\0"
    "linux-g++\0"
    "linux-g++\0"
#endif
;
#define QT_CONFIGURE_SETTINGS_PATH "/etc/xdg/qt5"
#define QT_CONFIGURE_LIBLOCATION_TO_PREFIX_PATH "../"
#define QT_CONFIGURE_HOSTBINDIR_TO_EXTPREFIX_PATH "../../../"
#define QT_CONFIGURE_HOSTBINDIR_TO_HOSTPREFIX_PATH "../../../"
#ifdef QT_BUILD_QMAKE
# define QT_CONFIGURE_SYSROOTIFY_PREFIX false
# define QT_CONFIGURE_CROSSBUILD false
#endif
#define QT_CONFIGURE_PREFIX_PATH qt_configure_prefix_path_str + 12
#ifdef QT_BUILD_QMAKE
# define QT_CONFIGURE_EXT_PREFIX_PATH qt_configure_ext_prefix_path_str + 12
# define QT_CONFIGURE_HOST_PREFIX_PATH qt_configure_host_prefix_path_str + 12
#endif
